/*
 * Segment 4 Datenübernahme und Aufbereitung, SWE 4.5 Abfrage Puffer
 * Copyright (C) 2007-2015 BitCtrl Systems GmbH
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.dua.abfrpuffer;

import java.util.Calendar;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.DataState;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.config.Aspect;
import de.bsvrz.dav.daf.main.config.ConfigurationObject;
import de.bsvrz.dav.daf.main.config.ObjectSet;
import de.bsvrz.dav.daf.main.config.ObjectTimeSpecification;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.dav.daf.main.config.SystemObjectType;
import de.bsvrz.sys.funclib.bitctrl.dua.DUAKonstanten;
import de.bsvrz.sys.funclib.dataIdentificationSettings.DataIdentification;
import de.bsvrz.sys.funclib.dataIdentificationSettings.UpdateListener;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Die Hauptklasse der Aggregationeinheit. Liest aus der Konfiguration die Liste
 * aller Langzetizaehlstellen, erzeugt uns stuert Instanzen von allen
 * benoetigten Aggregatoren, und beim jeden einkommenden Datensatz informiert
 * alle Agregatoren darueber, Verwaltet auch eine Abbildung von Fahrstreifen zu
 * assoziierten LangzeitZaehlstellen.
 *
 * @author BitCtrl Systems GmbH, Bachraty
 *
 */
public class AggregatorVerwaltung implements SBArchivAnfrageBearbeiter {

	/**
	 * Ermoeglicht die Felermeldungen zu speichern.
	 */
	private static final Debug LOGGER = Debug.getLogger();

	/**
	 * Konstanten zur Bearbeitung der Datensaetzen.
	 */
	private static final String ASP_AGREGINTTAG = "asp.aggregationsIntervallTag";
	private static final String TYP_LANGZEITZAEHLSTELLE = "typ.langZeitZählstelle";
	private static final String AUSWERTEQUERRSCHNITTE = "AuswerteQuerschnitte";
	private static final String MESSQUERSCHNITTE = "Messquerschnitt";
	private static final String ATG_AUSWERTEQUERSCHNITT = "atg.auswerteQuerschnitt";
	private static final String FAHRSTREIFEN = "FahrStreifen";

	/**
	 * Benutzt um die Zeitstempel aufrunden auf ganze Tage.
	 */
	private static Calendar calendar = Calendar.getInstance();
	/**
	 * Datenbeschreibung fuer die LZD von FS, die bearbeitet werden.
	 */
	private static DataDescription ddLveLzd;
	/**
	 * Verbindung zum DAV.
	 */
	private ClientDavInterface dav;
	/**
	 * Archiv Anfragen Generator.
	 */
	private StreamBasierteArchivAnfrage anfrage;
	/**
	 * List aller module, die Aggregierte Daten ausrechnen werden.
	 */
	private final LinkedList<AbstraktAggregator> aggregatoren;
	/**
	 * Map vom FS zur zuhoerigen LZZ.
	 */
	private HashMap<SystemObject, LZZWrapper> mapFSzumLZZ;
	/**
	 * Menge der LZZ.
	 */
	private Set<LZZWrapper> lzzMenge;
	/**
	 * Wird zur ArchivAnfragen benutzt, wenn man dem aeltesten ZS sucht.
	 */
	private boolean zeitStempelGefunden;

	/**
	 * ErzeugungsIntervall fuer LZD.
	 */
	private long lzdErzeugungsIntervall;
	/**
	 * Anzahl der angekommenen DatenSaetzen.
	 */
	private long anzahlDatenSaetze;

	/**
	 * Erwitert ein Fahrstreifen mit Information ueber den Anzahl der gekommenen
	 * Datensaetzen im letzten Zeitintervall.
	 *
	 * @author BitCtrl Systems GmbH, Bachraty
	 */
	protected class FSWrapper {

		/**
		 * Beginn des letzten Zeitintervalls.
		 */
		private long zeitStempelTAG;
		/**
		 * Anzahl der DatenSaetze.
		 */
		private long anzahlDatenSaetze;
		/**
		 * Anzahl der DS gestern.
		 */
		private long anzahlDatenSaetzeGestern;
		/**
		 * System Objekt das FS.
		 */
		private final SystemObject so;

		/**
		 * StandardKonstruktor.
		 *
		 * @param so
		 *            System Objekt, das FS
		 */
		public FSWrapper(final SystemObject so) {
			this.so = so;
			zeitStempelTAG = AbfrPufferKonstanten.ZEIT_BEGINN;
			anzahlDatenSaetze = 0;
		}

		/**
		 * Aktualisiert den Anzahl der gekommennen Datensaetzen im letzten
		 * zeitIntervall und bestimmt, ob aggregierte Daten berechnet werden
		 * sollen - wenn fuer alle FS der erste Datensatz gekommen ist, oder der
		 * vom ersten beliebigen FS der zweite DS gekommen ist.
		 *
		 * @param fsMenge
		 *            Menge der FS die zu dem selben FS gehoeren
		 * @param letzterZeitstempel
		 *            Letzter ZeitIntervall
		 * @return true, wenn die Aggregierte Daten berechnet werden sollen,
		 *         sonst false
		 */
		public boolean update(final Collection<FSWrapper> fsMenge, final long letzterZeitstempel) {

			// der erste DS ist gekommen
			if ((letzterZeitstempel > this.zeitStempelTAG)
					|| ((letzterZeitstempel == this.zeitStempelTAG) && (anzahlDatenSaetze == 0))) {

				this.zeitStempelTAG = letzterZeitstempel;
				anzahlDatenSaetzeGestern = anzahlDatenSaetze;
				anzahlDatenSaetze = 1;
				// alle haben schon wenigstens ein DS
				for (final FSWrapper fs : fsMenge) {
					if (!fs.istErsterDSgekommen(letzterZeitstempel)) {
						return false;
					}
				}
				return true;
			} else if (++anzahlDatenSaetze == 2) {
				boolean leerGefunden = false;
				// alle andere haben weniger als zwei DS
				// und wenigstens ein FS hat noch keinen DS
				for (final FSWrapper fs : fsMenge) {
					if (fs == this) {
						continue;
					}
					if (!fs.istMaxEinDSgekommen(letzterZeitstempel)) {
						return false;
					}
					if (fs.istNochKeinGekommen(letzterZeitstempel)) {
						leerGefunden = true;
					}
				}
				return leerGefunden;
			}
			return false;
		}

		/**
		 * Liefert den Anzahl der DS von gestern.
		 *
		 * @return Anzahl
		 */
		public long getAnzahlDSGestern() {
			return anzahlDatenSaetzeGestern;
		}

		/**
		 * Bestimmt, ob das FS schon dem ersten DS bekommen hat.
		 *
		 * @param zeitStempel
		 *            Zeitstempel des Tages
		 * @return true wenn alle FS mehr als 0 DS bekommen haben
		 */
		public boolean istErsterDSgekommen(final long zeitStempel) {
			if (zeitStempelTAG > zeitStempel) {
				return true;
			}
			if ((zeitStempelTAG == zeitStempel) && (anzahlDatenSaetze > 0)) {
				return true;
			}
			return false;
		}

		/**
		 * Bestimmt, ob das FS nicht mehr als ein DS bekommen hat.
		 *
		 * @param zeitStempel
		 *            Zeitstempel des Tages
		 * @return true wenn alle FS nicht mehr als 1 DS bekommen haben
		 */
		public boolean istMaxEinDSgekommen(final long zeitStempel) {
			if (zeitStempelTAG < zeitStempel) {
				return true;
			}
			if ((zeitStempelTAG == zeitStempel) && (anzahlDatenSaetze < 2)) {
				return true;
			}
			return false;
		}

		/**
		 * Bestimmt ob das FS noch keinen DS bekommen hat.
		 *
		 * @param zeitStempel
		 *            Zeitstempel des Tages
		 * @return true, wenn FS noch keinen DS bekommen hat
		 */
		public boolean istNochKeinGekommen(final long zeitStempel) {
			if (zeitStempelTAG < zeitStempel) {
				return true;
			}
			if ((zeitStempelTAG == zeitStempel) && (anzahlDatenSaetze == 0)) {
				return true;
			}
			return false;
		}

		/**
		 * Ergibt das SystemObjekt des FS.
		 *
		 * @return SystemObject FS
		 */
		public SystemObject getSo() {
			return so;
		}
	}

	/**
	 * Erweitert den LZZ SystemObjekt um die Menge der assoziierten FS.
	 *
	 * @author BitCtrl Systems GmbH, Bachraty
	 *
	 */
	protected class LZZWrapper {
		/**
		 * Das System Objekt der LZZ.
		 */
		private final SystemObject so;
		/**
		 * Mapt assoziierte FS SystemObjekte zum FSWrapper Objekten mit
		 * erweiterten Info.
		 */
		private final HashMap<SystemObject, FSWrapper> fsMap;
		/**
		 * Ob man schon aggregierte Daten erstellen kann.
		 */
		private boolean sollAggregieren;
		/**
		 * Neuste Tages-Zeitstempel von allen eingekommenen Daten.
		 */
		private long letzterTag = AbfrPufferKonstanten.ZEIT_BEGINN;
		/**
		 * Aeltester Tages-Zeitstempel von allen eingekommenen Daten.
		 */
		private long ersterTag = AbfrPufferKonstanten.ZEIT_BEGINN;

		/**
		 * Erfragt den letzten Zeitstepel von allen eingekommenen Daten.
		 *
		 * @return Zeitstempel
		 */
		public long getLetzterTag() {
			return letzterTag;
		}

		/**
		 * Stellt den aeltesten Zeitstempel fest.
		 *
		 * @param zeitStempelTag
		 *            Zeitstempel
		 */
		public void setErsterTag(final long zeitStempelTag) {
			ersterTag = zeitStempelTag;
		}

		/**
		 * Erfragt dem aeltesten Tages-Zeitstepel von allen eingekommenen Daten.
		 *
		 * @return Zeitstempel
		 */
		public long getErsterTag() {
			return ersterTag;
		}

		/**
		 * Erfragt, ob man aggregieren soll.
		 *
		 * @return true, wenn der Aggregationsflag true ist
		 */
		public boolean isSollAggregieren() {
			return sollAggregieren;
		}

		/**
		 * Stellt dem Aggregationsflag als false fest.
		 */
		public void resetSollAggregieren() {
			sollAggregieren = false;
		}

		/**
		 * Standardkonstruktor.
		 *
		 * @param so
		 *            Das SystemObjekt der LZZ
		 */
		public LZZWrapper(final SystemObject so) {
			this.so = so;
			fsMap = new HashMap<>();
		}

		/**
		 * Erfragt nach dem Anzahl der eingekommenen Datensaetzen von Gestern.
		 *
		 * @return Anzahl der DS
		 */
		public long getAnzahlDatenSaetzeGestern() {
			long sum = 0;
			for (final FSWrapper fs : fsMap.values()) {
				sum += fs.getAnzahlDSGestern();
			}
			return sum;
		}

		/**
		 * Aktualisiert den anzahl der einkommenen DS fuer das Fahrstreifen.
		 *
		 * @param soFs
		 *            Das Fahrstreifen
		 * @param zeitStempelTAG
		 *            Zeitstempell des Intervalls
		 * @return true, wenn die Aggregierte Daten berechnet werden sollen,
		 *         sonst false
		 */
		public boolean update(final SystemObject soFs, final long zeitStempelTAG) {
			final FSWrapper fs = fsMap.get(soFs);
			final boolean ret = fs.update(fsMap.values(), zeitStempelTAG);
			if (ret) {
				sollAggregieren = ret;
			}
			if ((letzterTag == AbfrPufferKonstanten.ZEIT_BEGINN) || (letzterTag < zeitStempelTAG)) {
				letzterTag = zeitStempelTAG;
			}

			if ((ersterTag == AbfrPufferKonstanten.ZEIT_BEGINN) || (ersterTag > zeitStempelTAG)) {
				ersterTag = zeitStempelTAG;
			}

			return ret;
		}

		/**
		 * Fuegt ein neues FS zu die LZZ.
		 *
		 * @param fsWrapp
		 *            FS Objekt
		 * @param systemObject
		 *            System Objekt des FS
		 */
		public void add(final FSWrapper fsWrapp, final SystemObject systemObject) {
			fsMap.put(systemObject, fsWrapp);
		}

		/**
		 * Ergibt das System Objekt der LZZ.
		 *
		 * @return das SystemObjekt
		 */
		public SystemObject getLzzSo() {
			return so;
		}

		/**
		 * Ergibt die Menge von FS als FSWrapper Objekte.
		 *
		 * @return fsMenge
		 */
		public Collection<FSWrapper> getFsMenge() {
			return fsMap.values();
		}

		/**
		 * Ergibt die Menge von FS als SystemObjekte Objekte.
		 *
		 * @return fsMenge
		 */
		public Collection<SystemObject> getFsSoMenge() {
			return fsMap.keySet();
		}
	}

	/**
	 * Standard Konstruktor.
	 */
	public AggregatorVerwaltung() {
		aggregatoren = new LinkedList<>();
	}

	/**
	 * Fuegt ein Aggregator in die Liste der verwalteten Agg. ein.
	 *
	 * @param aggreg
	 *            Eine konkrete Implementations des Aggregators
	 */
	public void addAggregator(final AbstraktAggregator aggreg) {
		aggregatoren.add(aggreg);
	}

	/**
	 * Init, erzeutgt die Liste der LZZ liest die letzten DS aus dem Archiv aus
	 * und nach dem synchronisiert die Aggregatoren, damit der letzte erzeugte
	 * DatenSatz aus dem selben zeitIntervall ist.
	 *
	 * @param connection
	 *            Verbindung zum DAV
	 * @param zeitPunktJetzt
	 *            Aktueller Zeitpunkt
	 */
	public void initialisiere(final ClientDavInterface connection, final long zeitPunktJetzt) {

		final Aspect asp = connection.getDataModel().getAspect(AggregatorVerwaltung.ASP_AGREGINTTAG);
		mapFSzumLZZ = new HashMap<>();
		dav = connection;

		final SystemObjectType lzzTyp = connection.getDataModel().getType(AggregatorVerwaltung.TYP_LANGZEITZAEHLSTELLE);
		final Collection<SystemObjectType> lzzTypListe = new LinkedList<>();
		lzzTypListe.add(lzzTyp);

		final Collection<SystemObject> alleLZZ = connection.getDataModel().getObjects(null, lzzTypListe,
				ObjectTimeSpecification.valid());
		for (final SystemObject lzz : alleLZZ) {
			final LZZWrapper lzzWrapperObjekt = new LZZWrapper(lzz);
			final ConfigurationObject coLzz = (ConfigurationObject) lzz;

			final ObjectSet aqInMenge = coLzz.getObjectSet(AggregatorVerwaltung.AUSWERTEQUERRSCHNITTE);

			final Set<SystemObject> aqMengeGesamt = new HashSet<>();
			if (aqInMenge == null) {
				AggregatorVerwaltung.LOGGER.warning("LZZ " + coLzz.getPid() + " entaelt keine Auswertequerschnitte");
			} else {
				aqMengeGesamt.addAll(aqInMenge.getElements());
			}
			if (aqMengeGesamt.size() == 0) {
				AggregatorVerwaltung.LOGGER.warning("LZZ " + coLzz.getPid() + " enthaelt keine Auswertequerschnitte");
			}

			final Data atgLangZeit = lzz
					.getConfigurationData(connection.getDataModel().getAttributeGroup("atg.langZeitZählstelle"));
			if (atgLangZeit != null) {
				if ((atgLangZeit.getItem("InRichtung").getReferenceValue("Auswertequerschnitt") != null) && (atgLangZeit
						.getItem("InRichtung").getReferenceValue("Auswertequerschnitt").getSystemObject() != null)) {
					aqMengeGesamt.add(atgLangZeit.getItem("InRichtung").getReferenceValue("Auswertequerschnitt")
							.getSystemObject());
				}
				if ((atgLangZeit.getItem("GegenRichtung").getReferenceValue("Auswertequerschnitt") != null)
						&& (atgLangZeit.getItem("GegenRichtung").getReferenceValue("Auswertequerschnitt")
								.getSystemObject() != null)) {
					aqMengeGesamt.add(atgLangZeit.getItem("GegenRichtung").getReferenceValue("Auswertequerschnitt")
							.getSystemObject());
				}
			}

			if (aqMengeGesamt.size() == 0) {
				AggregatorVerwaltung.LOGGER.warning("LZZ " + coLzz.getPid() + " entaelt keine Auswertequerschnitte");
				continue;
			}

			for (final SystemObject aq : aqMengeGesamt) {
				if (aq.isValid()) {
					final ConfigurationObject coAq = (ConfigurationObject) aq;
					final Data data = coAq.getConfigurationData(
							connection.getDataModel().getAttributeGroup(AggregatorVerwaltung.ATG_AUSWERTEQUERSCHNITT));
					if (data == null) {
						AggregatorVerwaltung.LOGGER.warning("AQ" + aq.getPid() + " "
								+ AggregatorVerwaltung.ATG_AUSWERTEQUERSCHNITT + " ist nicht konfigureiert");
						continue;
					}
					final String mqPid = data.getTextValue(AggregatorVerwaltung.MESSQUERSCHNITTE).getValueText();
					final SystemObject mq = connection.getDataModel().getObject(mqPid);
					if (mq == null) {
						AggregatorVerwaltung.LOGGER
						.warning("AQ" + aq.getPid() + " ist nicht mit messquerschnitt assoziiert");
						continue;
					}
					final ConfigurationObject coMq = (ConfigurationObject) mq;
					final ObjectSet fsMenge = coMq.getObjectSet(AggregatorVerwaltung.FAHRSTREIFEN);

					if (fsMenge == null) {
						AggregatorVerwaltung.LOGGER.warning("MQ " + mq.getPid() + " entaelt keine Fahrstreifen");
						continue;
					}
					final List<SystemObject> fsListe = fsMenge.getElements();
					if (fsListe.size() == 0) {
						AggregatorVerwaltung.LOGGER.warning("MQ " + mq.getPid() + " entaelt keine Fahrstreifen");
						continue;
					}

					for (final SystemObject fs : fsListe) {
						if (fs.isValid()) {
							lzzWrapperObjekt.add(new FSWrapper(fs), fs);
							mapFSzumLZZ.put(fs, lzzWrapperObjekt);
						}
					}
				}
			}
		}

		lzzMenge = new HashSet<>(mapFSzumLZZ.values());

		AggregatorVerwaltung.ddLveLzd = new DataDescription(
				connection.getDataModel().getAttributeGroup(DUAKonstanten.ATG_LZD),
				connection.getDataModel().getAspect(DUAKonstanten.ASP_EXTERNE_ERFASSUNG));

		AggregatorVerwaltung.LOGGER.info("LZZ Menge: " + lzzMenge);

		for (final AbstraktAggregator aggreg : aggregatoren) {
			aggreg.initialisiere(asp, connection, alleLZZ, zeitPunktJetzt);
		}

		anfrage = new StreamBasierteArchivAnfrage(dav);
		for (final LZZWrapper lzz : lzzMenge) {

			// Wir bestimmen, ob schon Aggregierte Daten im Archiv sind
			long zs = System.currentTimeMillis();
			for (final AbstraktAggregator agg : aggregatoren) {
				final long aggZs = agg.getLetzteAggregierteDatenZS(lzz.getLzzSo());
				if (aggZs < zs) {
					zs = aggZs;
				}
			}
			if (zs != AbfrPufferKonstanten.ZEIT_BEGINN) {
				continue;
			}

			// Wenn nicht, dan finden wir die aeletesten FS Daten
			zs = findeAeltestenZeitStempel(lzz.getFsSoMenge());

			if (zs == AbfrPufferKonstanten.ZEIT_BEGINN) {
				continue;
			}

			// Wir beginnen an diesem Tag, so der vorhaerige ist der letzte
			AggregatorVerwaltung.calendar.setTimeInMillis(zs);
			AggregatorVerwaltung.calendar.add(Calendar.DAY_OF_MONTH, -1);
			zs = AggregatorVerwaltung.calendar.getTimeInMillis();

			for (final AbstraktAggregator aggreg : aggregatoren) {
				aggreg.initLetzteAggregierteDatenZS(lzz.getLzzSo(), zs);
			}
		}

		// // Archiv einstellungen
		// DataDescription archivDatenBeschreibung = new DataDescription(
		// DAV.getDataModel().getAttributeGroup(AbfrPufferKonstanten.ATG_ARCHIV),
		// DAV.getDataModel().getAspect(AbfrPufferKonstanten.ASP_PARAMETERSOLL),
		// (short)0);
		//
		// archivSettingsMgr = new SettingsManager(DAV,
		// new
		// DataIdentification(DAV.getDataModel().getConfigurationAuthority(),
		// archivDatenBeschreibung ));

		// Callback-Methode: Wird vom DAV aufgerufen, wenn eine an/abzumeldende
		// DID identifiziert wurde
		AbfragePufferdaten.getArchivSettingsManager().addUpdateListener(new UpdateListener() {
			@Override
			public void update(final DataIdentification did, final Data alteDatei, final Data neueDatei) {
				for (final AbstraktAggregator agg : aggregatoren) {
					if ((alteDatei == null) && (neueDatei != null)) {
						agg.bearbeiteNeueEinstellung(did, neueDatei);
					} else if ((alteDatei != null) && (neueDatei == null)) { // Nicht
						// mehr
						// parametriert
						agg.bearbeiteGeloeschteEinstellung(did, alteDatei);
					} else if ((alteDatei != null) && (neueDatei != null)) { // Umparametriert
						agg.bearbeiteVeraenderteEinstellung(did, neueDatei);
					}
				}
			}
		});

		// Mehrfache Instanzierung fuehrt zu
		// archivSettingsMgr.start();

	}

	/**
	 * Wird beim jedem eingekommenen LZD Datensatz gerufen.
	 *
	 * @param resultData
	 *            ein empfangener Datensatz
	 */
	public void update(final ResultData resultData) {

		if (resultData.getData() == null) {
			return;
		}
		final LZZWrapper lzzWrapper = mapFSzumLZZ.get(resultData.getObject());
		if (lzzWrapper != null) {
			final long tag = AggregatorVerwaltung.getTagesBeginnVonZeitStempel(resultData.getDataTime());
			lzzWrapper.update(resultData.getObject(), tag);
		}

	}

	/**
	 * Erfragt dem Anzahl der eingekomenen DS fuer die LZZ, die das Fahrstreifen
	 * fs enthaelt.
	 *
	 * @param fs
	 *            Fahrstreifen
	 * @return Anzahl der DS
	 */
	public long getAnzahlDatenSaetzeLZZ(final SystemObject fs) {
		final LZZWrapper lzzWrapper = mapFSzumLZZ.get(fs);
		return lzzWrapper.getAnzahlDatenSaetzeGestern();
	}

	/**
	 * Schreibt die Anzahl aller DS fuer alle LZZ aus.
	 */
	public void printAnzahlDatenSaetzeLZZ() {
		for (final LZZWrapper lzz : lzzMenge) {
			System.out.println(" LZZ " + lzz.getLzzSo().getPid() + " " + lzz.getAnzahlDatenSaetzeGestern());
		}
	}

	/**
	 * Startet die Aggregation, wird am ende des Abrufes gerufen.
	 */
	public void trigger() {
		for (final LZZWrapper lzzWrapper : lzzMenge) {
			if (lzzWrapper.isSollAggregieren()) {

				final long letzterTag = lzzWrapper.getLetzterTag();
				final long ersterTag = lzzWrapper.getErsterTag();
				// Wir bestimmen, ob schon Aggregierte Daten im Archiv sind

				AggregatorVerwaltung.calendar.setTimeInMillis(ersterTag);
				AggregatorVerwaltung.calendar.add(Calendar.DAY_OF_MONTH, -1);
				for (final AbstraktAggregator agg : aggregatoren) {

					if (!agg.isLZZArchiviert(lzzWrapper.getLzzSo())) {
						continue;
					}
					final long aggZs = agg.getLetzteAggregierteDatenZS(lzzWrapper.getLzzSo());

					// Wenn wir keine ZS gefunden haben, dann initialisieren wir
					// die Aggregatoren
					if (aggZs == AbfrPufferKonstanten.ZEIT_BEGINN) {
						for (final AbstraktAggregator aggreg : aggregatoren) {
							aggreg.initLetzteAggregierteDatenZS(lzzWrapper.getLzzSo(),
									AggregatorVerwaltung.calendar.getTimeInMillis());
						}
					} else if (aggZs < AggregatorVerwaltung.calendar.getTimeInMillis()) {
						// Wenn wir eine ZS gefunden haben, dann beginnen wir am
						// naechsten Tag
						AggregatorVerwaltung.calendar.setTimeInMillis(aggZs);
					}
				}
				AggregatorVerwaltung.calendar.add(Calendar.DAY_OF_MONTH, 1);

				while (AggregatorVerwaltung.calendar.getTimeInMillis() < letzterTag) {
					final long abfrageBeginn = AggregatorVerwaltung.calendar.getTimeInMillis();
					AggregatorVerwaltung.calendar.add(Calendar.DAY_OF_MONTH, 1);
					berechneLangZeitdaten(lzzWrapper, abfrageBeginn, AggregatorVerwaltung.calendar.getTimeInMillis());
				}
				lzzWrapper.setErsterTag(AggregatorVerwaltung.calendar.getTimeInMillis());
				lzzWrapper.resetSollAggregieren();

			}
		}
	}

	/**
	 * Berechnet die LangZeitDaten, triggert alle Agregatoren.
	 *
	 * @param lzz
	 *            LZZ fuer die die Daten erzeugt werden sollen
	 * @param zeitStempelVon
	 *            Beginn ZeitStempel des ZeitIntervalls fuer dem die Daten
	 *            erzeugt werden sollen
	 * @param zeitStempelBis
	 *            Ende ZeitStempel des ZeitIntervalls fuer dem die Daten erzeugt
	 *            werden sollen
	 */
	public void berechneLangZeitdaten(final LZZWrapper lzz, final long zeitStempelVon, final long zeitStempelBis) {

		if (AbstraktAggregator.isKeinArchiviert(lzz.getLzzSo(), aggregatoren)) {
			return;
		}

		anfrage.generiereAnfrage(lzz.getFsSoMenge(), AggregatorVerwaltung.ddLveLzd, false, zeitStempelVon,
				zeitStempelBis - 1000);
		if (!anfrage.istErfolgreich()) {
			AggregatorVerwaltung.LOGGER.error("Fehler bei der Anfrage, FS daten koennen nicht bearbeitet werden");
			return;
		}
		anzahlDatenSaetze = 0;
		lzdErzeugungsIntervall = 0;
		anfrage.bearbeiteAntwortStream(this);
		boolean ermittelbar = ((zeitStempelBis - zeitStempelVon) * lzz.getFsSoMenge().size()) == (anzahlDatenSaetze
				* lzdErzeugungsIntervall);

		// Wenn es kein FS gibt, dann ist es nicht ermittelbar obwohl die Zahlen
		// ( alles gleich 0) stimmt
		if (lzz.getFsSoMenge().size() == 0) {
			ermittelbar = false;
		}

		for (final AbstraktAggregator aggreg : aggregatoren) {
			if (aggreg.isLZZArchiviert(lzz.getLzzSo())) {
				aggreg.erzeugeErgebniss(lzz.getLzzSo(), zeitStempelVon, zeitStempelBis, ermittelbar);
			}
		}
	}

	@Override
	public void bearbeiteArchivDaten(final SystemObject objekt, final DataDescription datenBeschreibung,
			final DataState status, final Data daten, final long zeitStempel) {
		if (daten == null) {
			return;
		}
		lzdErzeugungsIntervall = daten.getTimeValue("T").getMillis();
		anzahlDatenSaetze++;
		for (final AbstraktAggregator aggreg : aggregatoren) {
			final SystemObject lzz = mapFSzumLZZ.get(objekt).getLzzSo();
			if (aggreg.isLZZArchiviert(lzz)) {
				aggreg.bearbeiteDatenSatz(lzz, daten, zeitStempel);
			}
		}
	}

	/**
	 * Findet den aeltesten ZeitStempel fuer die Menge der FS in dem Archiv.
	 *
	 * @param fsMenge
	 *            Mnege der FS
	 * @return Zeitstempel
	 */
	private long findeAeltestenZeitStempel(final Collection<SystemObject> fsMenge) {
		final long jetzt = System.currentTimeMillis();

		final Calendar zeitpunkt = Calendar.getInstance();
		zeitpunkt.setTimeInMillis(AbfrPufferKonstanten.ZEIT_BEGINN);
		int i = 0;
		zeitStempelGefunden = false;

		do {
			zeitpunkt.add(Calendar.YEAR, 1);
			anfrage.generiereAnfrage(fsMenge, AggregatorVerwaltung.ddLveLzd, true, 1, zeitpunkt.getTimeInMillis());
			if (!anfrage.istErfolgreich()) {
				AggregatorVerwaltung.LOGGER.error("Letzter FS Datensatz koennte nicht gefunden werden");
				return AbfrPufferKonstanten.ZEIT_BEGINN;
			}
			anfrage.bearbeiteAntwortStream(new SBArchivAnfrageBearbeiter() {
				@Override
				public void bearbeiteArchivDaten(final SystemObject objekt, final DataDescription datenBeschreibung,
						final DataState status, final Data daten, final long zeitStempel) {
					if (status != DataState.END_OF_ARCHIVE) {
						AggregatorVerwaltung.this.setZeitStempelGefunden(true);
					}
				}
			});
		} while ((zeitpunkt.getTimeInMillis() < jetzt) && !zeitStempelGefunden);

		if (!zeitStempelGefunden) {
			return AbfrPufferKonstanten.ZEIT_BEGINN;
		}
		zeitpunkt.add(Calendar.YEAR, -1);

		zeitStempelGefunden = false;

		i = 0;
		do {
			zeitpunkt.add(Calendar.MONTH, 1);
			anfrage.generiereAnfrage(fsMenge, AggregatorVerwaltung.ddLveLzd, true, 1, zeitpunkt.getTimeInMillis());
			if (!anfrage.istErfolgreich()) {
				AggregatorVerwaltung.LOGGER.error("Letzter FS Datensatz koennte nicht gefunden werden");
				return AbfrPufferKonstanten.ZEIT_BEGINN;
			}
			anfrage.bearbeiteAntwortStream(new SBArchivAnfrageBearbeiter() {
				@Override
				public void bearbeiteArchivDaten(final SystemObject objekt, final DataDescription datenBeschreibung,
						final DataState status, final Data daten, final long zeitStempel) {
					if (status != DataState.END_OF_ARCHIVE) {
						AggregatorVerwaltung.this.setZeitStempelGefunden(true);
					}
				}
			});
		} while ((i++ < 12) && !zeitStempelGefunden);

		zeitpunkt.add(Calendar.MONTH, -1);

		i = 0;
		zeitStempelGefunden = false;
		do {
			zeitpunkt.add(Calendar.DAY_OF_MONTH, 1);
			anfrage.generiereAnfrage(fsMenge, AggregatorVerwaltung.ddLveLzd, true, 1, zeitpunkt.getTimeInMillis());
			if (!anfrage.istErfolgreich()) {
				AggregatorVerwaltung.LOGGER.error("Letzter FS Datensatz koennte nicht gefunden werden");
				return AbfrPufferKonstanten.ZEIT_BEGINN;
			}
			anfrage.bearbeiteAntwortStream(new SBArchivAnfrageBearbeiter() {
				@Override
				public void bearbeiteArchivDaten(final SystemObject objekt, final DataDescription datenBeschreibung,
						final DataState status, final Data daten, final long zeitStempel) {
					if (status != DataState.END_OF_ARCHIVE) {
						AggregatorVerwaltung.this.setZeitStempelGefunden(true);
					}
				}
			});
		} while ((i++ < 31) && !zeitStempelGefunden);
		zeitpunkt.add(Calendar.DAY_OF_MONTH, -1);
		return zeitpunkt.getTimeInMillis();
	}

	/**
	 * Setzt die Variable ein, wird als Wrapper aus Zugriffsgruenden in
	 * unbenannten Kinder-Klassen benutzt.
	 *
	 * @param zeitStempelGefunden
	 *            true, wenn ein Zeitstepel im Archiv gefunden ist
	 */
	public void setZeitStempelGefunden(final boolean zeitStempelGefunden) {
		this.zeitStempelGefunden = zeitStempelGefunden;
	}

	/**
	 * Aufrundet den ZeitStempel so, dass man dem Beginn des Tages bekommt in
	 * dem der Zeitstepel liegt.
	 *
	 * @param zeitStempel
	 *            Ein Zeitstempel
	 * @return Beginn des Tages, in dem der ZeitStempel liegt
	 */
	static long getTagesBeginnVonZeitStempel(final long zeitStempel) {
		AggregatorVerwaltung.calendar.setTimeInMillis(zeitStempel);
		AggregatorVerwaltung.calendar.set(Calendar.HOUR_OF_DAY, 0);
		AggregatorVerwaltung.calendar.set(Calendar.MINUTE, 0);
		AggregatorVerwaltung.calendar.set(Calendar.SECOND, 0);
		AggregatorVerwaltung.calendar.set(Calendar.MILLISECOND, 0);
		return AggregatorVerwaltung.calendar.getTimeInMillis();
	}
}
